package at.ac.uibk.gitsearch.service;

import static at.ac.uibk.gitsearch.web.rest.AccountResourceIT.TEST_USER_LOGIN;
import static org.hamcrest.CoreMatchers.containsStringIgnoringCase;
import static org.hamcrest.Matchers.anyOf;
import static org.hamcrest.Matchers.contains;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.everyItem;
import static org.hamcrest.Matchers.hasItemInArray;
import static org.hamcrest.Matchers.hasProperty;
import static org.hamcrest.Matchers.is;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.codeability.sharing.plugins.api.SharingPluginConfig;
import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO.ExerciseType;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.node.NodeValidationException;
import org.junit.Assert;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithMockUser;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestServerConfiguration;
import at.ac.uibk.gitsearch.service.PluginManagementService.ConnectorConfigWrapper;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;

@SpringBootTest(classes = GitsearchApp.class)
@WithMockUser(value = TEST_USER_LOGIN, authorities = "sharing")
public class SearchServiceIT {
	
	@Autowired
	RestHighLevelClient searchClient;
	
	@MockBean
	PluginManagementService pluginManagementService;

	private static  final Logger LOGGER = LoggerFactory.getLogger(SearchServiceIT.class);
	
	@Autowired
	private SearchService searchService;

	@BeforeAll
	public static void setUpESServer() throws IOException, NodeValidationException {
		ElasticSearchTestServerConfiguration.getTestInstance().startTestNode();
	}
	
	@BeforeEach
	public void initMock() {
		Collection<ConnectorConfigWrapper> mockConfigs = Collections.singletonList(
				new ConnectorConfigWrapper(new SharingPluginConfig("mockedPlugin", new SharingPluginConfig.Action[] {
						new SharingPluginConfig.Action("mockedAction", "http://unused/unused-transferURL", "mocked Action", "true")
				}), "http://unused/hopefully-unused")
				);
		when(pluginManagementService.getRegisteredPluginConfigs()).thenReturn(mockConfigs);
	}

	@Test
	public void testAllSearch() throws Exception {
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null,null);
		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);
		
		Assert.assertTrue("At least one hit?", searchResultPage.getSearchResult().size() >= 1);
		Assert.assertEquals("We start at 0", 0, searchResultPage.getPageStartIndex());
		LOGGER.info("found {} hits for all", searchResultPage.getHitCount());
	}

	@Test
	public void testFullTextSearch() throws Exception {
		LOGGER.info("starting testFullTextSearch");
		final String WIEN = "Wien";
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null,null);
		SearchInputDTO searchQuery = new SearchInputDTO(WIEN, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
		org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
		org.junit.Assert.assertThat(searchResultPage.getSearchResult(),
				everyItem(anyOf(hasProperty("metadata", hasProperty("description", containsString(WIEN))),
						hasProperty("metadata", hasProperty("title", containsString(WIEN)))
				// hasProperty("keywords", containsString("Wien"))
				)));

		LOGGER.info("exiting testFullTextSearch");
	}

	@Test
	public void testSearchByAutor() throws Exception {
		final String PODLIPNIG = "Stefan Podlipnig";
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, PODLIPNIG, null);
		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
		org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
		org.junit.Assert.assertThat(searchResultPage.getSearchResult(), everyItem(
				hasProperty("metadata", hasProperty("creator", hasItemInArray(hasProperty("name", containsString(PODLIPNIG)))))));

	}

	@Test
	public void testProgrammingLanguageSearch() throws Exception {
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO("JAVA", null, null, null, null,null);
		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		assertNotNull(searchResultPage.getSearchResult());
		assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
		assertThat(searchResultPage.getSearchResult(), everyItem(
				hasProperty("metadata", hasProperty("programmingLanguage", hasItemInArray(containsStringIgnoringCase("Java"))))));

	}

	@Test
	public void testCreatorSearch() throws IOException {

		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, "Podlipnig",null);
		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
		org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
		assertThat(searchResultPage.getSearchResult(), everyItem(hasProperty("metadata",
				hasProperty("creator", hasItemInArray(hasProperty("name", containsString("Podlipnig")))))));

	}

	@Test
	public void testKeywordSearch() throws IOException {
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null,null);
		org.codeability.sharing.plugins.api.search.SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
		org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
		assertThat(searchResultPage.getSearchResult(),
				everyItem(hasProperty("metadata", hasProperty("keyword", hasItemInArray(containsString("latex"))))));

	}

	@Test
	public void testFormatSearch() throws IOException {
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null,"latex");
		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
		org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
		assertThat(searchResultPage.getSearchResult(),
				everyItem(hasProperty("metadata", hasProperty("format", hasItemInArray(containsString("latex"))))));

	}

	@Test
	public void testTypesSearch() throws IOException {
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null,null);
		searchMetadata.setTypes(Collections.singletonList(ExerciseType.COLLECTION));
		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
		org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
		assertThat(searchResultPage.getSearchResult(),
				everyItem(hasProperty("metadata", hasProperty("keyword", hasItemInArray(containsString("latex"))))));

	}

	@Test
	public void testLanguagesSearch() throws IOException {
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null,null);
		searchMetadata.setNaturalLanguage(List.of("de"));;
		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
		org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
		assertThat(searchResultPage.getSearchResult(),
				everyItem(hasProperty("metadata", hasProperty("language", hasItemInArray(containsString("de"))))));

	}

	@Test
	public void testTypesSearchNegative() throws IOException {
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null,null);
		searchMetadata.setTypes(Collections.singletonList(ExerciseType.OTHER));
		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
		org.junit.Assert.assertTrue("We expect no hit for \"latex\" and ExerciseType.OTHER", searchResultPage.getHitCount() == 0);

	}

//	@Test()
//	@Ignore() // Test funktioniert momentan nicht?
	public void testLicenseSearch() throws IOException {
		
		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, "MIT", null,null);
		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
		org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
		assertThat(searchResultPage.getSearchResult(),
				everyItem(hasProperty("metadata", hasProperty("license", containsString("MIT")))));

	}
	
	@Test
	public void testKeywordAutocompletion() throws IOException {
		final List<AutoCompleteEntry> keywordsAutoComplete = searchService.getKeywordsAutoComplete("Jav", 10);
		assertThat(keywordsAutoComplete, contains(hasProperty("target", is("Java"))));
	}

	@Test
	public void testCreatorAutocompletion() throws IOException {
		final List<AutoCompleteEntry> creatorAutoComplete = searchService.getCreatorAutoComplete("Pod", 10);
		assertThat(creatorAutoComplete, contains(hasProperty("target", is("Stefan Podlipnig"))));
		final List<AutoCompleteEntry> creatorAutoComplete2 = searchService.getCreatorAutoComplete("Po", 10);
		assertThat(creatorAutoComplete2, contains(hasProperty("target", is("Stefan Podlipnig"))));
	}

	@Test
	public void testContributorAutocompletion() throws IOException {
		final List<AutoCompleteEntry> contributorAutoComplete = searchService.getContributorAutoComplete("Bast", 10);
		assertThat(contributorAutoComplete, contains(hasProperty("target", is("Daniel Bastta"))));
	}

	@Test
	public void testContributorCreatorAutocompletion() throws IOException {
		final List<AutoCompleteEntry> contributorAutoComplete = searchService.getContributorCreatorAutoComplete("Bast", 10);
		assertThat(contributorAutoComplete, contains(hasProperty("target", is("Daniel Bastta"))));
	}
}
